unit btree;
{****************************************************************************
*     BTREE FILE HANDLING  Version 1.02                                     *
*
*     Copyright   R. J. Betts  1995
****************************************************************************}
interface
uses dos;
const
  B_EOF              = 35;
  B_RECORD_NOT_FOUND = 36;
  B_DUPLICATE_KEY    = 37;
  B_KEY_NOT_FOUND    = 38;
  B_MEMORY_LOW       = 39;
  B_REWRITE_ERROR    = 40;
  B_WRITE_ERROR      = 41;
  B_INDEX_CORRUPT    = 42;
  B_INDEX_NOT_FOUND  = 43;

  B_PAGE_STACK : integer = 10;

  MaxKeyLength = 50;
  MaxDepth     = 5;

type
  B_KeyType = (B_BYTES,B_STRING,B_INTEGER,B_WORD);

  KeyStr   = array[1..MaxKeyLength] of byte;

  KeyItem = record
               p   : longint;  { key record pointer }
               r   : longint;  { data record pointer }
               Key : KeyStr;
             end;

  Key_Def = record
              KeyType : B_KeyType;
              Off     : integer;
              Len     : integer;
            end;

  KeyPath    = record
                 Page      : longint;
                 item      : integer;
               end;

  { Data file object }
  BFile = object
            F      : File;
            RecNo  : longint;
            RecSize: integer;

            constructor Init;
            function Create(Name:string;Size:integer):integer;
            function CloseFile:integer;
            function DeleteRecord:integer;
            function Lock(r:longint):integer;
            function Unlock(r:longint):integer;
            function Open(Name:string;Size:integer):integer;
            function Read(var Buf;r:longint):integer;
            function RewriteRecord(var Buf):integer;
            function Write(var Buf):integer;
          end;

  { Index data fie pair }
  BIndex = object (BFile)
             Data       : BFile;
             Keys       : integer;
             KeyDefs    : array[1..5] of Key_Def;
             KeyLen     : integer;
             KeySize    : integer;  { Size of key + pointers }
             RootPage   : longint;
             IndexError : integer;
             KeysPerPage: integer;
             LastPage   : pointer;
             LastEmpty  : integer;
             ID         : word;
             KeyPage    : longint;
             KeyPos     : integer;
             SearchDepth: integer;
             SearchPath : array[1..MaxDepth] of KeyPath;


             constructor Init;
             function CloseFile:integer;
             function Create(Name,DataName:string;Size:integer):integer;
             procedure DefineKey(key:integer;KeyType:B_KeyType;Off,Len:integer);
             function DeleteRecord(var Buf):integer;
             function Next(Var Buf):integer;
             function Open(Name,DataName:string;Size:integer):integer;
             function Prev(Var Buf):integer;
             function Read(var Buf):integer;
             function RewriteRecord(var Buf):integer;
             function Sort(Name,DataName:string;Size:integer):integer;
             function Start(var Buf):integer;
             function Write(var Buf):integer;

             function AddKey(Key:KeyItem):integer;
             procedure BuildKey(var Buf;var Key:KeyStr);
             function CreateIndexHeader:integer;
             function FindKey(var Key:KeyItem;var i:integer;var PagePtr:longint):boolean;
             function GetKeyPage(PagePtr:longint):pointer;
             procedure InvalidateStack;
             function SearchIndex(var Buf;var L:longint;SearchNext:boolean):boolean;
             function SetSearchPath(PagePtr:longint;i:integer):boolean;
             procedure SplitRoot(P:pointer);
             function UpdatePage(var Page;r:longint):integer;
             procedure UpdateRoot;
             function WritePage(var Page):longint;
           end;

  PIndex = ^BIndex;

function InitialiseBTREE:boolean;
procedure SetFileHandles(h:integer);

implementation

const KeySpaceSize = 2040;

type

  PKeyPage = ^BKeyPage;
  PStack   = ^Page_Stack;

  Page_Element = record
                  Handle : integer;
                  p      : PKeyPage;
               end;

  Page_Stack = array[1..99] of Page_Element;



{ Keys are stored as
array[0..pred(KeysPerPage)] of record
                                key record: longint
                                Record no : longint;
                                Key       : array[1..keylen] of byte
                               end;
}

   Key_Page = record
                ID       : integer;
                LastEmpty: longint;
                items    : integer;
                KeySpace : array[0..pred(KeySpaceSize)] of byte;
              end;

   Key_Header_Page = record
                       ID       : integer;
                       Keys     : integer;
                       KeyDefs  : array[1..5] of Key_Def;
                       DataName : string[64];
                       Root     : longint;
                       Empty    : longint;
                       filler   : array[0..1946] of byte;
                     end;



   BKeyPage = object
                Page   : Key_Page;
                tree   : PIndex;
                Parent : PKeyPage;
                Pidx   : integer;
                R      : longint;   { record pointer to this page }

                constructor Init;
                procedure add(var Key:KeyItem;i:integer);
                procedure Delete(i:integer);
                procedure DeleteItem(P:PKeyPage;i:integer);
                procedure remove(i,j:integer);
                function FindKey(var Key:KeyItem;var i:integer;var PagePtr:longint):boolean;
                function FindNext(var Key:KeyItem;var i:integer;var PagePtr:longint):boolean;
                function FindPrev(var Key:KeyItem;var i:integer;var PagePtr:longint):boolean;
                function Items:integer;
                function GetLeaf(i:integer):longint;
                procedure SetLeaf(i:integer;LeafPtr:longint);
                function GetRecord(i:integer):longint;
                function FirstLeaf(i:integer):pointer;
                function LastLeaf(i:integer):pointer;
                function UpdatePage:integer;
                procedure BalanceWithRight(Sib:PKeyPage);
                procedure BalanceWithLeft(Sib:PKeyPage);
                procedure SplitWithRight(Sib:PKeyPage);
                procedure SplitWithLeft(Sib:PKeyPage);
                procedure MergeWithLeft(Sib:PKeyPage);
                procedure MergeWithRight(Sib:PKeyPage);
                procedure GetKey(var Key:KeyItem;i:integer);
                procedure SetKey(var Key:KeyItem;i:integer);
                procedure Insert(var Key:KeyItem;i:integer);
                procedure InsertKeys(Var Keys;i,NoOfKeys:integer);
                procedure Empty;
                procedure PageFull(P:PKeyPage);
                procedure PageLow(P:PKeyPage);
              end;


var

  Node       : BKeyPage;
  Root       : PKeyPage;

  MaxStack,
  PageStack  : integer;
  pPageStack : PStack;

function Compare(var a;var b;Len:integer):integer;
var pA,pB : ^Char;
    i : integer;
begin
  Compare:=0;
  pA:=@A; pB:=@B;
  for i:=1 to Len do
  begin
    if pA^<pB^ then
    begin
      Compare:=-1;
      exit;
    end;
    if pA^>pB^ then
    begin
      Compare:=1;
      exit;
    end;
    inc(pA);
    inc(pB);
  end;
end;
function CompareByte(var a;b:byte;Len:integer):boolean;
var pA : ^byte;
    i : integer;
begin
  CompareByte:=true;
  pA:=@A;
  for i:=1 to Len do
  begin
    if pA^<>b then
    begin
      CompareByte:=false;
      exit;
    end;
    inc(pA);
  end;
end;

procedure SetFileHandles(h:integer);
{ set the number of file handles this program can use }
{ DOS defaults to 20 }
begin
  asm
     mov bx,h
     mov ah,67h
     int 21h
  end;
end;

function InitialiseBTREE:boolean;
var i : integer;
    p : PStack;
begin
  InitialiseBTREE:=false;
  MaxStack:=B_PAGE_STACK;
  PageStack:=0;
  GetMem(pPageStack,MaxStack*sizeof(Page_Element));
  if pPageStack=nil then
    exit;
  fillchar(pPageStack^,MaxStack*sizeof(Page_Element),0);
  p:=pPageStack;
  for i:=1 to MaxStack do
  begin
    p^[i].p:=new(PKeyPage,Init);
    if p^[i].p=nil then
      exit;
  end;
  InitialiseBTREE:=true;
end;
(*///////////////  Key page methods //////////////////////*)
constructor BKeyPage.Init;
begin
  fillchar(Page,sizeof(Page),0);
  R:=0;
  Parent:=nil;
  tree:=nil;
end;

procedure BKeyPage.Add(var Key:KeyItem;i:integer);
var Off : integer;
    Len : integer;
begin
  if Page.Items<tree^.KeysPerPage then
  begin
    Len:=(Page.Items-i)*(tree^.KeySize)+4;
    Off:=(tree^.KeySize)*i;
    move(Page.Keyspace[Off],Page.KeySpace[Off+tree^.KeySize],Len);
    move(Key.Key,Page.KeySpace[Off+8],tree^.KeyLen);
    move(Key.r,Page.KeySpace[Off+4],4);
    move(Key.p,Page.KeySpace[Off],4);
    inc(Page.Items);
    if (Page.Items=tree^.KeysPerPage) then
      if Parent=nil then
        tree^.SplitRoot(@Self)
      else
        Parent^.PageFull(@Self)
    else
      if UpdatePage<>0 then;
  end;
end;

procedure BKeyPage.Delete(i:integer);
var L : Longint;
    LowWaterMark : integer;
    LowLeaf  : PKeyPage;
    Key   : KeyItem;
begin
  LowWaterMark:=tree^.KeysPerPage div 2;
  L:=GetLeaf(succ(i));
  if L=0 then
  begin
    remove(i,1);
    if (Items<LowWaterMark) and (Parent<>nil) then
      Parent^.PageLow(@Self)
    else
      if UpdatePage<>0 then;
    exit;
  end;
  LowLeaf:=FirstLeaf(succ(i));
  if (LowLeaf=nil) then
  begin
    exit;
  end;
  LowLeaf^.GetKey(Key,0);
  Key.p:=GetLeaf(succ(i));
  SetKey(Key,i);
  LowLeaf^.DeleteItem(@Self,0);
  if UpdatePage<>0 then;
end;

function BKeyPage.FirstLeaf(i:integer):pointer;
var L : Longint;
    P : PKeyPage;
begin
  FirstLeaf:=@Self;
  L:=GetLeaf(i);
  if L>0 then
  begin
    P:=tree^.GetKeyPage(L);
    if P=nil then
    begin
      FirstLeaf:=nil;
      exit;
    end;
    P^.Parent:=@Self;
    if not tree^.SetSearchPath(L,i) then exit;
    FirstLeaf:=P^.FirstLeaf(0);
  end;
end;

function BKeyPage.LastLeaf(i:integer):pointer;
var L : Longint;
    P : PKeyPage;
begin
  LastLeaf:=@Self;
  L:=GetLeaf(i);
  if L>0 then
  begin
    P:=tree^.GetKeyPage(L);
    if P=nil then
    begin
      LastLeaf:=nil;
      exit;
    end;
    P^.Parent:=@Self;
    if not tree^.SetSearchPath(L,i) then exit;
    LastLeaf:=P^.LastLeaf(P^.items);
  end;
end;

procedure BKeyPage.MergeWithRight(Sib:PKeyPage);
var Key : KeyItem;
begin
  Parent^.GetKey(Key,pidx);
  Key.p:=GetLeaf(items);
  Insert(Key,Items);
  InsertKeys(Sib^.Page.KeySpace,Sib^.Items,Items);
  SetLeaf(items,Sib^.GetLeaf(Sib^.Items));
  Sib^.Remove(0,Sib^.Items);
  if UpdatePage=0 then
  begin
    if Sib^.UpdatePage<>0 then;
  end;
  Parent^.DeleteItem(@Self,pidx);
end;

procedure BKeyPage.MergeWithLeft(Sib:PKeyPage);
var Key : KeyItem;
begin
  Parent^.GetKey(Key,pidx);
  Key.p:=GetLeaf(Sib^.Items);
  Sib^.Insert(Key,Sib^.Items);
  Sib^.InsertKeys(Page.KeySpace,Sib^.Items,Items);
  Sib^.SetLeaf(Sib^.items,GetLeaf(Items));
  Remove(0,Items);
  if Sib^.UpdatePage=0 then
    if UpdatePage=0 then;
  Parent^.DeleteItem(Sib,pidx);
end;

procedure BKeyPage.Empty;
begin
  fillchar(Page,sizeof(Page),0);
end;
function BKeyPage.FindKey(var Key:KeyItem;var i:integer;var PagePtr:longint):boolean;
var off : integer;
    j,k,m : integer;
    L     : longint;
begin
  FindKey:=false;
  Off:=8+i*tree^.KeySize;
  m:=i;
  for k:=m to pred(Page.Items) do
  begin
    i:=k;
    j:=Compare(Key.Key,Page.KeySpace[off],tree^.KeyLen);
    if j<0 then
    begin
      move(Page.KeySpace[Off-8],L,4);
      if L>0 then
      begin
        PagePtr:=L;
        FindKey:=tree^.FindKey(Key,i,PagePtr);
      end
      else
        move(Page.KeySpace[Off-4],Key.r,4);
      exit;
    end;
    if j=0 then
    begin
      FindKey:=true;
      move(Page.KeySpace[Off-4],Key.r,4);
      exit;
    end;
    Off:=Off+tree^.KeySize;
  end;
  inc(i);
  move(Page.KeySpace[Off-8],L,4);
  if L>0 then
  begin
    PagePtr:=L;
    FindKey:=tree^.FindKey(Key,i,PagePtr);
  end
  else
    Key.r:=-1;
end;
function BKeyPage.FindPrev(var Key:KeyItem;var i:integer;var PagePtr:longint):boolean;
var off : integer;
    j,k,m : integer;
    L     : longint;
    LowLeaf : PKeyPage;
begin
  FindPrev:=false;
  Off:=i*tree^.KeySize;
  move(Page.KeySpace[Off],L,4);
  if L>0 then
  begin
    LowLeaf:=LastLeaf(i);
    if LowLeaf<>nil then
    begin
      i:=Pred(LowLeaf^.Items);
      LowLeaf^.GetKey(Key,i);
      PagePtr:=LowLeaf^.R;
      FindPrev:=true;
    end;
    exit;
  end;
  if i=0 then
    Key.r:=-1
  else begin
    dec(i);
    GetKey(Key,i);
    FindPrev:=true;
  end;
end;
function BKeyPage.FindNext(var Key:KeyItem;var i:integer;var PagePtr:longint):boolean;
var off : integer;
    j,k,m : integer;
    L     : longint;
    LowLeaf : PKeyPage;
begin
  FindNext:=false;
  inc(i);
  Off:=i*tree^.KeySize;
  move(Page.KeySpace[Off],L,4);
  if L>0 then
  begin
    LowLeaf:=FirstLeaf(i);
    if LowLeaf<>nil then
    begin
      i:=0;
      LowLeaf^.GetKey(Key,i);
      PagePtr:=LowLeaf^.R;
      FindNext:=true;
    end;
    exit;
  end;
  if i>Pred(Items) then
    Key.r:=-1
  else begin
    GetKey(Key,i);
    FindNext:=true;
  end;
end;

procedure BKeyPage.GetKey(var Key:KeyItem;i:integer);
var Off : integer;
begin
  Off:=i*tree^.KeySize+8;
  move(Page.KeySpace[Off],Key.Key,tree^.KeyLen);
  move(Page.KeySpace[Off-4],Key.r,4);
  move(Page.KeySpace[Off-8],Key.p,4);
end;
procedure BKeyPage.SetKey(var Key:KeyItem;i:integer);
var Off : integer;
begin
  Off:=i*tree^.KeySize+8;
  move(Key.Key,Page.KeySpace[Off],tree^.KeyLen);
  move(Key.r,Page.KeySpace[Off-4],4);
  move(Key.p,Page.KeySpace[Off-8],4);
end;

function BKeyPage.Items:integer;
begin
  items:=Page.Items;
end;

function BKeyPage.UpdatePage:integer;
begin
  if (items=0) and (R<>1) then
  begin
    Page.LastEmpty:=tree^.LastEmpty;
    tree^.LastEmpty:=R;
  end;
  if (R=1) then
  begin
    Page.LastEmpty:=tree^.LastEmpty;
    inc(Page.ID);
    tree^.ID:=Page.ID;
  end;
  UpdatePage:=tree^.UpdatePage(Page,R);
end;

function BKeyPage.GetLeaf(i:integer):longint;
var off : integer;
      L : longint;
begin
  if i<0 then
    GetLeaf:=0
  else
  begin
    Off:=i*tree^.KeySize;
    move(Page.KeySpace[Off],L,4);
    GetLeaf:=L;
  end;
end;
function BKeyPage.GetRecord(i:integer):longint;
var off : integer;
      L : longint;
begin
  if i<0 then
    GetRecord:=-1
  else
  begin
    Off:=i*tree^.KeySize+4;
    move(Page.KeySpace[Off],L,4);
    GetRecord:=L;
  end;
end;

procedure BKeyPage.SetLeaf(i:integer;LeafPtr:longint);
var off : integer;
begin
  Off:=i*tree^.KeySize;
  move(LeafPtr,Page.KeySpace[Off],4);
end;

procedure BKeyPage.Insert(var Key:KeyItem;i:integer);
var Off : integer;
    Len : integer;
begin
  Len:=(Page.Items-i)*(tree^.KeySize)+4;
  Off:=(tree^.KeySize)*i;
  move(Page.Keyspace[Off],Page.KeySpace[Off+tree^.KeySize],Len);
  move(Key.Key,Page.KeySpace[Off+8],tree^.KeyLen);
  move(Key.r,Page.KeySpace[Off+4],4);
  move(Key.p,Page.KeySpace[Off],4);
  inc(Page.Items);
end;

procedure BKeyPage.InsertKeys(var Keys;i,NoOfKeys:integer);
var src,dest : integer;
    Size: integer;
begin
  src:=i*tree^.KeySize;
  dest:=(i+NoOfKeys)*tree^.KeySize;
  Size:=(Items-i)*tree^.KeySize;
  move(Page.KeySpace[src],Page.KeySpace[Dest],Size);
  move(Keys,Page.KeySpace[src],NoOfKeys*tree^.KeySize);
  inc(Page.Items,NoOfKeys);
end;

procedure BKeyPage.Remove(i,j:integer);
var src,dest,size : integer;
begin
  if (i<Page.Items) then
  begin
    dest:=i*tree^.KeySize;
    src:=(i+j)*tree^.KeySize;
    size:=(Page.Items-i-j+1)*tree^.KeySize;
    move(Page.KeySpace[src],Page.KeySpace[dest],size);
    dec(Page.Items,j);
    dest:=(Page.Items*tree^.KeySize)+4;
    size:=KeySpaceSize-dest;
    fillchar(Page.KeySpace[dest],size,0);
  end;
end;
procedure BKeyPage.DeleteItem(P:PKeyPage;i:integer);
var L : longint;
begin
  L:=GetLeaf(i);
  Remove(i,1);
  if Items<tree^.KeysPerPage div 2 then
    if Parent<>nil then
    begin
      Parent^.PageLow(@Self);
      exit;
    end;
  if (Parent=nil) and (Items=0) then
  begin
    Page:=P^.Page;
    P^.Page.Items:=0;
    if P^.UpdatePage<>0 then;
  end;
  if UpdatePage<>0 then;
end;
procedure BKeyPage.BalanceWithRight(Sib:PKeyPage);
{ Move some keys to the sibling to the right }
var i,j :integer;
    Key :KeyItem;
    src,dest : integer;
    NoOfKeys : integer;
begin
  i:=Sib^.Items;
  j:=(Page.Items+i+1) div 2;
  NoOfKeys:=j-i;
  Parent^.GetKey(Key,pidx);
  Key.p:=GetLeaf(Page.Items);
  Sib^.Insert(Key,0);
  dec(NoOfKeys);
  src:=(Page.Items-NoOfKeys)*tree^.KeySize;
  Sib^.InsertKeys(Page.KeySpace[src],0,NoOfKeys);
  Remove(Page.Items-NoOfKeys,NoOfKeys);
  GetKey(Key,pred(Page.Items));
  Remove(pred(Page.Items),1);
  SetLeaf(Page.Items,Key.p);
  Key.p:=R;
  Parent^.SetKey(Key,pidx);
  if (UpdatePage=0) then
    if Sib^.UpdatePage=0 then
      if Parent^.UpdatePage=0 then;
end;
procedure BKeyPage.BalanceWithLeft(Sib:PKeyPage);
{ Move some keys to the sibling to the left }
var i,j :integer;
    Key :KeyItem;
    src,dest : integer;
    NoOfKeys : integer;
begin
  i:=Sib^.Items;
  j:=(Page.Items+i+1) div 2;
  NoOfKeys:=j-i;
  Parent^.GetKey(Key,pred(pidx));
  Key.p:=Sib^.GetLeaf(Sib^.Items);
  Sib^.Insert(Key,i);
  Sib^.InsertKeys(Page.KeySpace,Sib^.Items,NoOfKeys);
  Remove(0,NoOfKeys);
  Sib^.GetKey(Key,pred(Sib^.Items));
  Sib^.Remove(pred(Sib^.Items),1);
  Sib^.SetLeaf(Sib^.Items,Key.p);
  Key.p:=Sib^.R;
  Parent^.SetKey(Key,pred(pidx));
  if (UpdatePage=0) then
    if Sib^.UpdatePage=0 then
      if Parent^.UpdatePage=0 then;
end;
procedure BKeyPage.SplitWithRight(Sib:PKeyPage);
{ Create new page taking some keys from this and right sib }
var i,j :integer;
    Key :KeyItem;
    src,dest : integer;
    NoOfKeys : integer;
    pPage    : PKeyPage;
begin
  pPage:=tree^.GetKeyPage(0);
  pPage^.Parent:=Parent;
  i:=Sib^.Items+Page.Items+1;
  j:=i div 3;

  Parent^.GetKey(Key,pidx);
  { Move keys from this }
  NoOfKeys:=Items-j;
  src:=(Items-NoOfKeys)*tree^.KeySize;
  pPage^.InsertKeys(Page.KeySpace[src],0,NoOfKeys);
  Key.P:=GetLeaf(Page.Items);
  Remove(Page.Items-NoOfKeys,NoOfKeys);
  pPage^.Insert(Key,NoOfKeys);

  { Move keys from the right sibling }
  NoOfKeys:=Sib^.Items-j;
  dest:=pPage^.Items*tree^.KeySize;
  pPage^.InsertKeys(Sib^.Page.KeySpace,pPage^.Items,NoOfKeys);
  Sib^.Remove(0,NoOfKeys);

  { set the parent key to the last one on this page }
  GetKey(Key,pred(Items));
  Remove(Pred(Items),1);
  SetLeaf(Page.Items,Key.p);
  Key.p:=R;
  Parent^.SetKey(Key,pidx);

  { Take the last key from the new page and add it to the parent}
  pPage^.GetKey(Key,pred(pPage^.Items));
  pPage^.Remove(pred(pPage^.Items),1);
  pPage^.SetLeaf(pPage^.Items,Key.p);
  pPage^.R:=tree^.WritePage(pPage^.Page);
  Key.p:=pPage^.R;
  if pPage^.R<0 then
  begin
    exit;
  end;
  if UpdatePage=0 then
    if Sib^.UpdatePage=0 then
      Parent^.Add(Key,succ(pidx));
end;

procedure BKeyPage.SplitWithLeft(Sib:PKeyPage);
{ Create new page taking some keys from this and sib }
var i,j :integer;
    Key :KeyItem;
    src,dest : integer;
    NoOfKeys : integer;
    pPage    : PKeyPage;
    Leaf     : longint;
begin
  pPage:=tree^.GetKeyPage(0);
  pPage^.Parent:=Parent;
  i:=Sib^.Items+Page.Items+1;
  j:=i div 3;

  Parent^.GetKey(Key,pred(pidx));
  { Move keys from the left sibling }
  NoOfKeys:=Sib^.Items-j;
  src:=(Sib^.Items-NoOfKeys)*tree^.KeySize;
  pPage^.InsertKeys(Sib^.Page.KeySpace[src],0,NoOfKeys);
  Sib^.Remove(Sib^.Items-NoOfKeys,NoOfKeys);
  Key.P:=Sib^.GetLeaf(Sib^.Page.Items);
  pPage^.Insert(Key,NoOfKeys);

  { Move keys from this }
  NoOfKeys:=Items-j;
  dest:=pPage^.Items*tree^.KeySize;
  pPage^.InsertKeys(Page.KeySpace,pPage^.Items,NoOfKeys);
  Remove(0,NoOfKeys);

  { set the parent key to the last one on sib page }
  Sib^.GetKey(Key,pred(Sib^.Items));
  Sib^.Remove(Pred(Sib^.Items),1);
  Key.p:=Sib^.R;
  Parent^.SetKey(Key,pred(pidx));

  { Take the last key from the new page and add it to the parent}
  pPage^.GetKey(Key,pred(pPage^.Items));
  pPage^.Remove(pred(pPage^.Items),1);
  pPage^.SetLeaf(pPage^.Items,Key.P);
  pPage^.R:=tree^.WritePage(pPage^.Page);
  if pPage^.R<0 then
  begin
    exit;
  end;
  Key.p:=pPage^.R;
  if UpdatePage=0 then
    if Sib^.UpdatePage=0 then
      Parent^.Add(Key,pidx);
end;
procedure BKeyPage.PageFull(P:PKeyPage);
{ The key page pointed at by P is full }
{ Attempt to balance it with a sibling to the right or left }
{ If this fails, split the page with a sibling }
var LPtr,RPtr : longint;
    pPage : PKeyPage;
    i     : integer;
begin
  i:=P^.Pidx;
  RPtr:=GetLeaf(Succ(i));
  if RPtr>0 then
  begin
    pPage:=tree^.GetKeyPage(RPtr);
    if pPage=nil then
      exit;
    pPage^.Parent:=@Self;
    if pPage^.Items<pred(tree^.KeysPerPage) then
    begin
      P^.BalanceWithRight(pPage);
      exit;
    end;
  end;
  LPtr:=GetLeaf(pred(i));
  if LPtr>0 then
  begin
    pPage:=tree^.GetKeyPage(LPtr);
    if pPage=nil then
      exit;
    pPage^.Parent:=@Self;
    if pPage^.Items<pred(tree^.KeysPerPage) then
    begin
      P^.BalanceWithLeft(pPage);
      exit;
    end;
  end;
  { One or both siblings must be full - choose one to split with }
  if (LPtr>0) then
  begin
    P^.SplitWithLeft(pPage);
  end
  else
    P^.SplitWithRight(pPage);
end;

procedure BKeyPage.PageLow(P:PKeyPage);
{ The page pointed at by P contains too few keys                   }
{ Attempt to balance it with the right or left sibling, otherwise  }
{ merge it with a sibling                                          }
var Right,Left : longint;
    i : integer;
    pPage : PKeyPage;
begin
  i:=P^.pidx;
  Right:=GetLeaf(succ(i));
  if (Right>0) then
  begin
    pPage:=tree^.GetKeyPage(Right);
    if pPage=nil then
      exit;
    pPage^.Parent:=@Self;
    if (pPage^.Items>tree^.KeysPerPage div 2) then
    begin
      BalanceWithRight(pPage);
      exit;
    end;
  end;
  Left:=GetLeaf(pred(i));
  if (Left>0) then
  begin
    pPage:=tree^.GetKeyPage(Right);
    if pPage=nil then
      exit;
    pPage^.Parent:=@Self;
    if (pPage^.Items>tree^.KeysPerPage div 2) then
    begin
      BalanceWithLeft(pPage);
      exit;
    end;
  end;
  if (Left>0) then
  begin
    P^.MergeWithLeft(pPage);
  end
  else
    P^.MergeWithRight(pPage);
end;

{//////  INDEX ROUTINES ///////}
{$i-}  { Error checking off }
constructor BIndex.Init;
begin
  RecSize:=KeySpaceSize+8;
  Data.Init;
  RootPage:=1;
  Keys:=0;
end;
function BIndex.AddKey(Key:KeyItem):integer;
var rtn : integer;
    PagePtr : longint;
    i       : integer;
    NewRoot : boolean;
    P       : PKeyPage;
    L       : longint;
begin
  AddKey:=0;
  SearchDepth:=0;
  LastPage:=nil;
  PagePtr:=1;
  i:=0;
  L:=Key.r;
  if FindKey(Key,i,PagePtr) then
   AddKey:=B_DUPLICATE_KEY
  else
  begin
    P:=LastPage;
    NewRoot:=false;
    if (IndexError=B_RECORD_NOT_FOUND) and (PagePtr=1) then
    begin
      NewRoot:=true;
      IndexError:=0;
      i:=0;
      P:=GetKeyPage(0);
      if P<>nil then
      begin
        P^.Empty;
        p^.Parent:=nil;
        LastEmpty:=0;
        P^.R:=WritePage(P^.Page);
        root:=P;
      end;
    end;
    if IndexError<>0 then
    begin
      AddKey:=IndexError;
      exit;
    end;
    LastEmpty:=P^.Page.LastEmpty;
    ID:=P^.Page.ID;
    Key.r:=L;
    P^.Add(Key,i);
    UpdateRoot;
  end;
end;
procedure BIndex.BuildKey(var Buf;var Key:KeyStr);
var i,j,n : integer;
    Off : integer;
    bb  : array[0..2047] of byte absolute Buf;
    b   : byte;
begin
  j:=1;
  for i:=1 to Keys do
  begin
    Off:=KeyDefs[i].Off;
    if KeyDefs[i].KeyType=B_STRING then
      inc(Off);
    move(bb[Off],Key[j],KeyDefs[i].Len);
    case (KeyDefs[i].KeyType) of
      B_STRING :
      begin
        b:=bb[pred(Off)];
        if b<KeyDefs[i].Len then
          fillchar(Key[j+b],KeyDefs[i].Len-b,0);
      end;
      B_INTEGER,
      B_WORD:
      begin
        if KeyDefs[i].Len>1 then
        begin
          b:=Key[j];
          Key[j]:=Key[succ(j)];
          Key[succ(j)]:=b;
        end;
        if KeyDefs[i].Len>2 then
        begin
          b:=Key[j];
          Key[j]:=Key[j+3];
          Key[j+3]:=b;
          b:=Key[succ(j)];
          Key[succ(j)]:=Key[j+2];
          Key[j+2]:=b;
        end;
        if KeyDefs[i].KeyType=B_INTEGER then
        begin
          if Key[j] and 128>0 then
            Key[j]:=Key[j] xor 128
          else
            Key[j]:=Key[j] or 128;
        end;
      end;
    end;
    j:=j+KeyDefs[i].Len;
  end;
end;
function BIndex.CloseFile:integer;
begin
  InvalidateStack;
  Close(F);
  CloseFile:=Data.CloseFile;
end;
function BIndex.Create(Name,DataName:string;Size:integer):integer;
var rtn : integer;
    i   : integer;
begin
  rtn:=Data.Create(DataName,Size);
  if rtn=0 then
  begin
    Assign(F,Name);
    rewrite(F,RecSize);
    rtn:=IOResult;
    if rtn=0 then
    begin
      rtn:=CreateIndexHeader;
    end;
  end;
  KeyLen:=0;
  KeyPos:=0;
  KeyPage:=0;
  for i:=1 to Keys do
    KeyLen:=KeyDefs[i].Len+KeyLen;
  KeySize:=KeyLen+8;
  KeysPerPage:=(KeySpaceSize) div (KeySize);
  if (KeysPerPage and 1=0) then dec(KeysPerPage);
  Create:=rtn;
end;
function BIndex.CreateIndexHeader;
var p : ^Key_Header_Page;
    i : integer;
begin
  fillchar(Node.Page,sizeof(Node.Page),0);
  p:=@Node.Page;
  p^.Keys:=Keys;
  for i:=1 to 5 do
  begin
    p^.KeyDefs[i].Off:=KeyDefs[i].Off;
    p^.KeyDefs[i].Len:=KeyDefs[i].Len;
  end;
  CreateIndexHeader:=BFile.Write(Node.Page);
end;

procedure BIndex.DefineKey(key:integer;KeyType:B_KeyType;Off,Len:integer);
begin
  if (key<1) or (key>5) or (Len>MaxKeyLength) then
     exit;
  KeyDefs[Key].KeyType:=KeyType;
  KeyDefs[Key].Off:=Off;
  KeyDefs[Key].Len:=Len;
  Keys:=Key;
end;


function BIndex.DeleteRecord(Var Buf):integer;
var Key : KeyItem;
    rtn : integer;
    i   : integer;
    Leaf : longint;
    P    : PKeyPage;
begin
  IndexError:=0;
  BuildKey(Buf,Key.Key);
  SearchDepth:=0;
  i:=0;
  Leaf:=1;
  LastPage:=nil;
  if not FindKey(Key,i,Leaf) then
    DeleteRecord:=B_KEY_NOT_FOUND
  else
  begin
    P:=LastPage;
    LastEmpty:=P^.Page.LastEmpty;
    ID:=P^.Page.ID;
    P^.Delete(i);
    rtn:=IndexError;
    if rtn=0 then
    begin
      Data.RecNo:=Key.r;
      rtn:=Data.DeleteRecord;
    end;
    DeleteRecord:=rtn;
  end;
end;
function BIndex.FindKey(var Key:KeyItem;var i:integer;var PagePtr:longint):boolean;
var Depth : integer;
    item  : integer;
    P     : pKeyPage;
begin
  FindKey:=false;
  if not SetSearchPath(PagePtr,i) then exit;
  P:=GetKeyPage(PagePtr);
  if (P=nil) then
    exit;
  if LastPage=nil then
    root:=P;
  P^.Parent:=LastPage;
  LastPage:=P;
  P^.pidx:=i;
  if SearchDepth>1 then i:=0;
  FindKey:=P^.FindKey(Key,i,PagePtr);
end;
function BIndex.GetKeyPage(PagePtr:longint):pointer;
var p : PStack;
    i,j : integer;
    pe  : Page_Element;
    ThisFile : FileRec absolute F;
begin
  { Search the page stack first }
  p:=pPageStack;
  i:=0;
  if PagePtr>0 then
  begin
    while (i<PageStack) do
    begin
      inc(i);
      if (p^[i].p<>nil) and (p^[i].handle=ThisFile.handle) and
         (p^[i].p^.r=PagePtr) then
      begin
        GetKeyPage:=p^[i].p;
        if (i>1) then      { Move this page to the top of the stack }
        begin
          pe:=p^[i];
          for j:=i downto 2 do
            p^[j]:=p^[pred(j)];
          p^[1]:=pe;
        end;
        exit;
      end;
    end;
  end;
  if PageStack<MaxStack then
  begin
    inc(PageStack);
    i:=PageStack;
  end
  else
  begin
    i:=MaxStack;
  end;
  p:=pPageStack;
  if PagePtr>0 then
    IndexError:=BFile.Read(p^[i].p^.Page,PagePtr)
  else
  begin
    p^[i].p^.Empty;
    IndexError:=0;
  end;
  if IndexError=0 then
  begin
    p^[i].Handle:=ThisFile.Handle;
    p^[i].p^.r:=PagePtr;
    p^[i].p^.tree:=@self;
    GetKeyPage:=p^[i].p;
    if (i>1) then      { Move this page to the top of the stack }
    begin
      pe:=p^[i];
      for j:=i downto 2 do
        p^[j]:=p^[pred(j)];
      p^[1]:=pe;
    end;
  end
  else
  begin
    p^[i].Handle:=0;
    GetKeyPage:=nil;
  end;
end;
procedure BIndex.InvalidateStack;
var i : integer;
    Index : FileRec absolute F;
begin
  for i:=1 to PageStack do
  begin
    if pPageStack^[i].Handle=Index.Handle then
      pPageStack^[i].Handle:=0;
  end;
end;


function BIndex.Next(var Buf):integer;
var rtn:integer;
    L  : longint;
begin
  if SearchIndex(Buf,L,true) then;
  if L<0 then
    rtn:=B_EOF
  else
  begin
    rtn:=Data.Read(Buf,L);
  end;
  Next:=rtn;
end;
function BIndex.Open(Name,DataName:string;Size:integer):integer;
var rtn : integer;
    i   : integer;
begin
{$i-}
  Assign(F,Name);
  rtn:=Data.Open(DataName,Size);
  if (rtn=0) then
  begin
    reset(F,RecSize);
    rtn:=IOResult;
    if rtn<>0 then
    begin
      if rtn=2 then
        rtn:=B_INDEX_NOT_FOUND;
      if Data.CloseFile=0 then;
    end;
  end;
  KeyLen:=0;
  KeyPos:=0;
  KeyPage:=0;
  for i:=1 to Keys do
    KeyLen:=KeyDefs[i].Len+KeyLen;
  KeySize:=KeyLen+8;
  KeysPerPage:=(KeySpaceSize) div (KeySize);
  if (KeysPerPage and 1=0) then dec(KeysPerPage);
  IF rtn=0 then
  begin
    if BFile.Read(Node,0)<>0 then
    begin
      rtn:=CreateIndexHeader;
    end;
  end;
  Open:=rtn;
end;

function BIndex.Prev(var Buf):integer;
var rtn:integer;
    L  : longint;
begin
  if SearchIndex(Buf,L,false) then;
  if L<0 then
    rtn:=B_EOF
  else
  begin
    rtn:=Data.Read(Buf,L);
  end;
  Prev:=rtn;
end;
function BIndex.Read(Var Buf):integer;
var RecsRead : integer;
    Key      : KeyItem;
    PagePtr  : longint;
    i        : integer;
begin
  IndexError:=0;
  BuildKey(Buf,Key.Key);
  SearchDepth:=0;
  PagePtr:=1;
  i:=0;
  if not FindKey(Key,i,PagePtr) then
  begin
    if IndexError>0 then Read:=IndexError else Read:=B_KEY_NOT_FOUND;
  end
  else
  begin
    Read:=Data.Read(Buf,Key.r);
  end;
  KeyPos:=i;
  KeyPage:=PagePtr;
end;
function BIndex.RewriteRecord(Var Buf):integer;
var Recs : word;
begin
  RewriteRecord:=Data.RewriteRecord(Buf);
end;

function BIndex.SearchIndex(var Buf;var L:longint;SearchNext:boolean):boolean;
var PagePtr : longint;
    Key     : KeyItem;
    i       : integer;
    P       : PKeyPage;
    found   : boolean;
    Depth   : integer;
begin
  Depth:=SearchDepth;
  PagePtr:=KeyPage;
  i:=KeyPos;
  IndexError:=0;
  BuildKey(Buf,Key.Key);
  LastPage:=nil;
  if KeyPage=0 then
  begin
    PagePtr:=1;
    SearchIndex:=FindKey(Key,i,PagePtr);
  end
  else
  begin
    P:=GetKeyPage(PagePtr);
    if P=nil then
      exit;
    P^.Parent:=LastPage;
    LastPage:=P;
    if SearchNext then
       found:=P^.FindNext(Key,i,PagePtr)
    else
      found:=P^.FindPrev(Key,i,PagePtr);
    if not found then
    begin
      while not found and (depth>1) do
      begin
        dec(Depth);
        i:=SearchPath[Depth].item;
        PagePtr:=SearchPath[Depth].Page;
        P:=GetKeyPage(PagePtr);
        if P=nil then
           exit;
        if SearchNext then
        begin
          if i<=pred(P^.Items) then
          begin
            Found:=true;
            P^.GetKey(Key,i);
            inc(SearchPath[depth].item);
          end
        end
        else begin
          if (i>0) then
          begin
            dec(i);
            P^.GetKey(Key,i);
            Found:=true;
            dec(SearchPath[depth].item);
          end;
        end;
      end;
      SearchDepth:=Depth;
    end;
  end;
  KeyPage:=PagePtr;
  KeyPos:=i;
  L:=Key.r;
end;

function BIndex.SetSearchPath(PagePtr:longint;i:integer):boolean;
begin
  if SearchDepth>0 then
    SearchPath[SearchDepth].item:=i;
  inc(SearchDepth);
  if SearchDepth>MaxDepth then
  begin
    IndexError:=B_INDEX_CORRUPT;
    SetSearchPath:=false;
    exit;
  end;
  SearchPath[SearchDepth].Page:=PagePtr;
  SetSearchPath:=true;
end;
function BIndex.Sort(Name,DataName:string;Size:integer):integer;
var Rtn  : integer;
    pBuf : pointer;
    L    : longint;
    Key  : KeyItem;
    Count : integer;
begin
  Sort:=0;
  rtn:=Data.Open(DataName,Size);
  if rtn<>0 then
  begin
    Sort:=Rtn;
    exit;
  end;
  rewrite(F,RecSize);
  rtn:=IOResult;
  Count:=0;
  if rtn=0 then
  begin
    rtn:=CreateIndexHeader;
    IF rtn=0 then
    begin
      GetMem(pBuf,Data.Recsize);
      if pBuf=nil then
        Sort:=B_MEMORY_LOW
      else
      begin
        L:=0;
        rtn:=Data.Read(pBuf^,L);
        while (rtn=0) do
        begin
          if not CompareByte(pBuf^,$ff,Data.RecSize) then
          begin
            BuildKey(pBuf^,Key.Key);
            Key.r:=L;
            Key.p:=0;
            rtn:=AddKey(Key);
          end;
          inc(L);
          if (rtn=0) or (rtn=B_DUPLICATE_KEY) then
            rtn:=Data.Read(pBuf^,L);
        end;
      end;
    end;
  end;
end;
procedure BIndex.SplitRoot(P:pointer);
var i : integer;
    Off : integer;
    LSib,RSib : longint;
    pLeaf     : PKeyPage;
    pRoot     : PKeyPage;
    L         : longint;
    Key       : KeyItem;
    RootID    : integer;
    RootEmpty : longint;
begin
  pLeaf:=P;
  pRoot:=GetKeyPage(0);
  i:=KeysPerPage div 2;
  pRoot^.InsertKeys(pLeaf^.Page.KeySpace,0,i);
  L:=pLeaf^.GetLeaf(i);
  pRoot^.SetLeaf(i,L);
  pRoot^.R:=WritePage(pRoot^.Page);
  if IndexError<>0 then exit;
  LSib:=pRoot^.R;

  Off:=succ(i)*KeySize;
  pRoot^.Empty;
  pRoot^.InsertKeys(pLeaf^.page.KeySpace[Off],0,i);
  L:=pLeaf^.GetLeaf(pLeaf^.Items);
  pRoot^.SetLeaf(i,L);
  pRoot^.R:=WritePage(pRoot^.Page);
  pRoot^.Parent:=P;
  if IndexError<>0 then exit;
  RSib:=pRoot^.R;

  pLeaf^.GetKey(Key,i);
  RootID:=pLeaf^.Page.ID;
  RootEmpty:=pLeaf^.Page.LastEmpty;
  pLeaf^.Empty;
  pLeaf^.Page.ID:=RootID;
  pLeaf^.Page.LastEmpty:=RootEmpty;
  pLeaf^.Insert(Key,0);
  pLeaf^.SetLeaf(0,LSib);
  pLeaf^.SetLeaf(1,RSib);
  IndexError:=pLeaf^.UpdatePage;

end;
function BIndex.Start(var Buf):integer;
var rtn:integer;
    L  : longint;
begin
  KeyPage:=0;
  KeyPos:=0;
  SearchDepth:=0;
  if SearchIndex(Buf,L,true) then;
  if L<0 then
    rtn:=B_EOF
  else
  begin
    rtn:=Data.Read(Buf,L);
  end;
  Start:=rtn;
end;
function BIndex.UpdatePage(var Page;r:longint):integer;
begin
  RecNo:=r;
  IndexError:=BFile.RewriteRecord(Page);
  UpdatePage:=IndexError;
end;
procedure BIndex.UpdateRoot;
begin
(*  if (root^.Page.LastEmpty<>LastEmpty) or (root^.Page.ID<>ID) then
  begin
    if root^.UpdatePage<>0 then;
  end;
*)
end;
function BIndex.Write(Var Buf):integer;
var Recs : word;
    Key  : KeyItem;
    rtn : integer;
begin
  IndexError:=0;
  BuildKey(Buf,Key.Key);
  rtn:=Data.Write(Buf);
  if rtn=0 then
  begin
    Key.p:=0;
    Key.r:=Data.RecNo;
    rtn:=AddKey(Key);
  end;
  Write:=rtn;
end;
function BIndex.WritePage(var Page):longint;
var err : integer;
    P   : ^Key_Page;
begin
  if LastEmpty<>0 then
  begin
    IndexError:=BFile.Read(Page,LastEmpty);
    P:=@Page;
    if P^.Items<>0 then
      WritePage:=-1
    else
    begin
      WritePage:=LastEmpty;
      LastEmpty:=P^.LastEmpty;
    end;
    exit;
  end;
  IndexError:=BFile.Write(Page);
  if IndexError=0 then
    WritePage:=pred(FileSize(F))
  else
    WritePage:=-1;
end;
(*///////////  Data file handling ///////////////////////*)
constructor BFile.Init;
begin
end;

function BFile.Open(Name:string;Size:integer):integer;
begin
  RecSize:=Size;
  Assign(F,Name);
  reset(F,RecSize);
  Open:=IOResult;
end;

function BFile.Create(Name:string;Size:integer):integer;
begin
  RecSize:=Size;
  Assign(F,Name);
  rewrite(F,RecSize);
  Create:=IOResult;
end;

function BFile.CloseFile:integer;
begin
  Close(F);
  CloseFile:=IOResult;
end;

function BFile.Lock(r:longint):integer;
begin
  Lock:=0; { No file locks for single user version }
end;
function BFile.Unlock(r:longint):integer;
begin
  Unlock:=0; { No file locks for single user version }
end;

function BFile.Read(Var Buf;r:longint):integer;
var RecsRead : integer;
begin
  RecNo:=r;
  Seek(F,r);
  BlockRead(F,Buf,1,RecsRead);
  r:=IOResult;
  if (RecsRead<1) and (r=0) then
    Read:=B_RECORD_NOT_FOUND
  else
    Read:=r;
end;

function BFile.DeleteRecord:integer;
var pRec : pointer;
    Recs : word;
    error: integer;
begin
  GetMem(pRec,RecSize);
  if pRec=NIL then
    DeleteRecord:=B_MEMORY_LOW
  else
  begin
    fillchar(pRec^,RecSize,$ff);
    Seek(F,RecNo);
    BlockWrite(F,pRec^,1,Recs);
    error:=IOResult;
    if (Recs<1) and (Error=0) then
      DeleteRecord:=B_REWRITE_ERROR
    else
      DeleteRecord:=error;
  end;
end;

function BFile.Write(Var Buf):integer;
var Recs : word;
    Error: integer;
begin
  Seek(F,FileSize(F));
  BlockWrite(F,Buf,1,Recs);
  Error:=IOResult;
  if (Recs<1) and (Error=0) then
    Write:=B_WRITE_ERROR
  else begin
    RecNo:=pred(FileSize(F));
    Write:=Error;
  end;
end;

function BFile.RewriteRecord(Var Buf):integer;
var Recs : word;
    Error : integer;
begin
  Seek(F,RecNo);
  BlockWrite(F,Buf,1,Recs);
  Error:=IOResult;
  if (Recs<1) and (Error=0) then
    RewriteRecord:=B_REWRITE_ERROR
  else
    RewriteRecord:=Error;
end;

end.